# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import proto


def test_basic_map():
    class Foo(proto.Message):
        tags = proto.MapField(proto.STRING, proto.STRING, number=1)

    foo = Foo(tags={"a": "foo", "b": "bar"})
    assert foo.tags["a"] == "foo"
    assert foo.tags["b"] == "bar"
    assert "c" not in foo.tags


def test_basic_map_with_underscore_field_name():
    class Foo(proto.Message):
        tag_labels = proto.MapField(proto.STRING, proto.STRING, number=1)

    foo = Foo(tag_labels={"a": "foo", "b": "bar"})
    assert foo.tag_labels["a"] == "foo"
    assert foo.tag_labels["b"] == "bar"
    assert "c" not in foo.tag_labels


def test_basic_map_assignment():
    class Foo(proto.Message):
        tags = proto.MapField(proto.STRING, proto.STRING, number=1)

    foo = Foo(tags={"a": "foo"})
    foo.tags["b"] = "bar"
    assert len(foo.tags) == 2
    assert foo.tags["a"] == "foo"
    assert foo.tags["b"] == "bar"
    assert "c" not in foo.tags


def test_basic_map_deletion():
    class Foo(proto.Message):
        tags = proto.MapField(proto.STRING, proto.STRING, number=1)

    foo = Foo(tags={"a": "foo", "b": "bar"})
    del foo.tags["b"]
    assert len(foo.tags) == 1
    assert foo.tags["a"] == "foo"
    assert "b" not in foo.tags



# Copyright 2017, Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import calendar
import datetime

import pytest
import pytz

from proto import datetime_helpers
from google.protobuf import timestamp_pb2


ONE_MINUTE_IN_MICROSECONDS = 60 * 1e6


def test_from_microseconds():
    five_mins_from_epoch_in_microseconds = 5 * ONE_MINUTE_IN_MICROSECONDS
    five_mins_from_epoch_datetime = datetime.datetime(
        1970, 1, 1, 0, 5, 0, tzinfo=datetime.timezone.utc
    )

    result = datetime_helpers._from_microseconds(five_mins_from_epoch_in_microseconds)

    assert result == five_mins_from_epoch_datetime


def test_to_rfc3339():
    value = datetime.datetime(2016, 4, 5, 13, 30, 0)
    expected = "2016-04-05T13:30:00.000000Z"
    assert datetime_helpers._to_rfc3339(value) == expected


def test_to_rfc3339_with_utc():
    value = datetime.datetime(2016, 4, 5, 13, 30, 0, tzinfo=datetime.timezone.utc)
    expected = "2016-04-05T13:30:00.000000Z"
    assert datetime_helpers._to_rfc3339(value, ignore_zone=False) == expected


def test_to_rfc3339_with_non_utc():
    zone = pytz.FixedOffset(-60)
    value = datetime.datetime(2016, 4, 5, 13, 30, 0, tzinfo=zone)
    expected = "2016-04-05T14:30:00.000000Z"
    assert datetime_helpers._to_rfc3339(value, ignore_zone=False) == expected


def test_to_rfc3339_with_non_utc_ignore_zone():
    zone = pytz.FixedOffset(-60)
    value = datetime.datetime(2016, 4, 5, 13, 30, 0, tzinfo=zone)
    expected = "2016-04-05T13:30:00.000000Z"
    assert datetime_helpers._to_rfc3339(value, ignore_zone=True) == expected


def test_ctor_wo_nanos():
    stamp = datetime_helpers.DatetimeWithNanoseconds(2016, 12, 20, 21, 13, 47, 123456)
    assert stamp.year == 2016
    assert stamp.month == 12
    assert stamp.day == 20
    assert stamp.hour == 21
    assert stamp.minute == 13
    assert stamp.second == 47
    assert stamp.microsecond == 123456
    assert stamp.nanosecond == 123456000


def test_ctor_w_nanos():
    stamp = datetime_helpers.DatetimeWithNanoseconds(
        2016, 12, 20, 21, 13, 47, nanosecond=123456789
    )
    assert stamp.year == 2016
    assert stamp.month == 12
    assert stamp.day == 20
    assert stamp.hour == 21
    assert stamp.minute == 13
    assert stamp.second == 47
    assert stamp.microsecond == 123456
    assert stamp.nanosecond == 123456789


def test_ctor_w_micros_positional_and_nanos():
    with pytest.raises(TypeError):
        datetime_helpers.DatetimeWithNanoseconds(
            2016, 12, 20, 21, 13, 47, 123456, nanosecond=123456789
        )


def test_ctor_w_micros_keyword_and_nanos():
    with pytest.raises(TypeError):
        datetime_helpers.DatetimeWithNanoseconds(
            2016, 12, 20, 21, 13, 47, microsecond=123456, nanosecond=123456789
        )


def test_rfc3339_wo_nanos():
    stamp = datetime_helpers.DatetimeWithNanoseconds(2016, 12, 20, 21, 13, 47, 123456)
    assert stamp.rfc3339() == "2016-12-20T21:13:47.123456Z"


def test_rfc3339_wo_nanos_w_leading_zero():
    stamp = datetime_helpers.DatetimeWithNanoseconds(2016, 12, 20, 21, 13, 47, 1234)
    assert stamp.rfc3339() == "2016-12-20T21:13:47.001234Z"


def test_rfc3339_w_nanos():
    stamp = datetime_helpers.DatetimeWithNanoseconds(
        2016, 12, 20, 21, 13, 47, nanosecond=123456789
    )
    assert stamp.rfc3339() == "2016-12-20T21:13:47.123456789Z"


def test_rfc3339_w_nanos_w_leading_zero():
    stamp = datetime_helpers.DatetimeWithNanoseconds(
        2016, 12, 20, 21, 13, 47, nanosecond=1234567
    )
    assert stamp.rfc3339() == "2016-12-20T21:13:47.001234567Z"


def test_rfc3339_w_nanos_no_trailing_zeroes():
    stamp = datetime_helpers.DatetimeWithNanoseconds(
        2016, 12, 20, 21, 13, 47, nanosecond=100000000
    )
    assert stamp.rfc3339() == "2016-12-20T21:13:47.1Z"


def test_rfc3339_w_nanos_w_leading_zero_and_no_trailing_zeros():
    stamp = datetime_helpers.DatetimeWithNanoseconds(
        2016, 12, 20, 21, 13, 47, nanosecond=1234500
    )
    assert stamp.rfc3339() == "2016-12-20T21:13:47.0012345Z"


def test_from_rfc3339_w_invalid():
    stamp = "2016-12-20T21:13:47"
    with pytest.raises(ValueError):
        datetime_helpers.DatetimeWithNanoseconds.from_rfc3339(stamp)


def test_from_rfc3339_wo_fraction():
    timestamp = "2016-12-20T21:13:47Z"
    expected = datetime_helpers.DatetimeWithNanoseconds(
        2016, 12, 20, 21, 13, 47, tzinfo=datetime.timezone.utc
    )
    stamp = datetime_helpers.DatetimeWithNanoseconds.from_rfc3339(timestamp)
    assert stamp == expected


def test_from_rfc3339_w_partial_precision():
    timestamp = "2016-12-20T21:13:47.1Z"
    expected = datetime_helpers.DatetimeWithNanoseconds(
        2016, 12, 20, 21, 13, 47, microsecond=100000, tzinfo=datetime.timezone.utc
    )
    stamp = datetime_helpers.DatetimeWithNanoseconds.from_rfc3339(timestamp)
    assert stamp == expected


def test_from_rfc3339_w_full_precision():
    timestamp = "2016-12-20T21:13:47.123456789Z"
    expected = datetime_helpers.DatetimeWithNanoseconds(
        2016, 12, 20, 21, 13, 47, nanosecond=123456789, tzinfo=datetime.timezone.utc
    )
    stamp = datetime_helpers.DatetimeWithNanoseconds.from_rfc3339(timestamp)
    assert stamp == expected


@staticmethod
@pytest.mark.parametrize(
    "fractional, nanos",
    [
        ("12345678", 123456780),
        ("1234567", 123456700),
        ("123456", 123456000),
        ("12345", 123450000),
        ("1234", 123400000),
        ("123", 123000000),
        ("12", 120000000),
        ("1", 100000000),
    ],
)
def test_from_rfc3339_test_nanoseconds(fractional, nanos):
    value = "2009-12-17T12:44:32.{}Z".format(fractional)
    assert (
        datetime_helpers.DatetimeWithNanoseconds.from_rfc3339(value).nanosecond == nanos
    )


def test_timestamp_pb_wo_nanos_naive():
    stamp = datetime_helpers.DatetimeWithNanoseconds(2016, 12, 20, 21, 13, 47, 123456)
    delta = stamp.replace(tzinfo=datetime.timezone.utc) - datetime_helpers._UTC_EPOCH
    seconds = int(delta.total_seconds())
    nanos = 123456000
    timestamp = timestamp_pb2.Timestamp(seconds=seconds, nanos=nanos)
    assert stamp.timestamp_pb() == timestamp


def test_timestamp_pb_w_nanos():
    stamp = datetime_helpers.DatetimeWithNanoseconds(
        2016, 12, 20, 21, 13, 47, nanosecond=123456789, tzinfo=datetime.timezone.utc
    )
    delta = stamp - datetime_helpers._UTC_EPOCH
    timestamp = timestamp_pb2.Timestamp(
        seconds=int(delta.total_seconds()), nanos=123456789
    )
    assert stamp.timestamp_pb() == timestamp


def test_from_timestamp_pb_wo_nanos():
    when = datetime.datetime(
        2016, 12, 20, 21, 13, 47, 123456, tzinfo=datetime.timezone.utc
    )
    delta = when - datetime_helpers._UTC_EPOCH
    seconds = int(delta.total_seconds())
    timestamp = timestamp_pb2.Timestamp(seconds=seconds)

    stamp = datetime_helpers.DatetimeWithNanoseconds.from_timestamp_pb(timestamp)

    assert _to_seconds(when) == _to_seconds(stamp)
    assert stamp.microsecond == 0
    assert stamp.nanosecond == 0
    assert stamp.tzinfo == datetime.timezone.utc


def test_replace():
    stamp = datetime_helpers.DatetimeWithNanoseconds(
        2016, 12, 20, 21, 13, 47, 123456, tzinfo=datetime.timezone.utc
    )

    # ns and ms provided raises
    with pytest.raises(TypeError):
        stamp.replace(microsecond=1, nanosecond=0)

    # No Nanoseconds or Microseconds
    new_stamp = stamp.replace(year=2015)
    assert new_stamp.year == 2015
    assert new_stamp.microsecond == 123456
    assert new_stamp.nanosecond == 123456000

    # Nanos
    new_stamp = stamp.replace(nanosecond=789123)
    assert new_stamp.microsecond == 789
    assert new_stamp.nanosecond == 789123

    # Micros
    new_stamp = stamp.replace(microsecond=456)
    assert new_stamp.microsecond == 456
    assert new_stamp.nanosecond == 456000

    # assert _to_seconds(when) == _to_seconds(stamp)
    # assert stamp.microsecond == 0
    # assert stamp.nanosecond == 0
    # assert stamp.tzinfo == datetime.timezone.utc


def test_from_timestamp_pb_w_nanos():
    when = datetime.datetime(
        2016, 12, 20, 21, 13, 47, 123456, tzinfo=datetime.timezone.utc
    )
    delta = when - datetime_helpers._UTC_EPOCH
    seconds = int(delta.total_seconds())
    timestamp = timestamp_pb2.Timestamp(seconds=seconds, nanos=123456789)

    stamp = datetime_helpers.DatetimeWithNanoseconds.from_timestamp_pb(timestamp)

    assert _to_seconds(when) == _to_seconds(stamp)
    assert stamp.microsecond == 123456
    assert stamp.nanosecond == 123456789
    assert stamp.tzinfo == datetime.timezone.utc


def _to_seconds(value):
    """Convert a datetime to seconds since the unix epoch.

    Args:
        value (datetime.datetime): The datetime to convert.

    Returns:
        int: Microseconds since the unix epoch.
    """
    assert value.tzinfo is datetime.timezone.utc
    return calendar.timegm(value.timetuple())



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .enums import Enum
from .fields import Field
from .fields import MapField
from .fields import RepeatedField
from .marshal import Marshal
from .message import Message
from .modules import define_module as module
from .primitives import ProtoType


DOUBLE = ProtoType.DOUBLE
FLOAT = ProtoType.FLOAT
INT64 = ProtoType.INT64
UINT64 = ProtoType.UINT64
INT32 = ProtoType.INT32
FIXED64 = ProtoType.FIXED64
FIXED32 = ProtoType.FIXED32
BOOL = ProtoType.BOOL
STRING = ProtoType.STRING
MESSAGE = ProtoType.MESSAGE
BYTES = ProtoType.BYTES
UINT32 = ProtoType.UINT32
ENUM = ProtoType.ENUM
SFIXED32 = ProtoType.SFIXED32
SFIXED64 = ProtoType.SFIXED64
SINT32 = ProtoType.SINT32
SINT64 = ProtoType.SINT64


__all__ = (
    "Enum",
    "Field",
    "MapField",
    "RepeatedField",
    "Marshal",
    "Message",
    "module",
    # Expose the types directly.
    "DOUBLE",
    "FLOAT",
    "INT64",
    "UINT64",
    "INT32",
    "FIXED64",
    "FIXED32",
    "BOOL",
    "STRING",
    "MESSAGE",
    "BYTES",
    "UINT32",
    "ENUM",
    "SFIXED32",
    "SFIXED64",
    "SINT32",
    "SINT64",
)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import proto


def test_int_init():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)
        baz = proto.Field(proto.INT32, number=2)

    foo = Foo(bar=42)
    assert foo.bar == 42
    assert foo.baz == 0
    assert not foo.baz
    assert Foo.pb(foo).bar == 42
    assert Foo.pb(foo).baz == 0


def test_int_rmw():
    class Foo(proto.Message):
        spam = proto.Field(proto.INT32, number=1)
        eggs = proto.Field(proto.INT32, number=2)

    foo = Foo(spam=42)
    foo.eggs = 76  # trombones led the big parade...
    assert foo.spam == 42
    assert foo.eggs == 76
    assert Foo.pb(foo).spam == 42
    assert Foo.pb(foo).eggs == 76
    foo.spam = 144
    assert foo.spam == 144
    assert foo.eggs == 76
    assert Foo.pb(foo).spam == 144
    assert Foo.pb(foo).eggs == 76


def test_int_del():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    foo = Foo(bar=42)
    assert foo.bar == 42
    del foo.bar
    assert foo.bar == 0
    assert not foo.bar


def test_int_size():
    class Foo(proto.Message):
        small = proto.Field(proto.INT32, number=1)
        big = proto.Field(proto.INT64, number=2)

    foo = Foo()
    foo.big = 2 ** 40
    assert foo.big == 2 ** 40
    with pytest.raises(ValueError):
        foo.small = 2 ** 40
    with pytest.raises(ValueError):
        Foo(small=2 ** 40)


def test_int_unsigned():
    class Foo(proto.Message):
        signed = proto.Field(proto.INT32, number=1)
        unsigned = proto.Field(proto.UINT32, number=2)

    foo = Foo()
    foo.signed = -10
    assert foo.signed == -10
    with pytest.raises(ValueError):
        foo.unsigned = -10
    with pytest.raises(ValueError):
        Foo(unsigned=-10)


def test_field_descriptor_idempotent():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1)

    bar_field = Foo.meta.fields["bar"]
    assert bar_field.descriptor is bar_field.descriptor


def test_int64_dict_round_trip():
    # When converting a message to other types, protobuf turns int64 fields
    # into decimal coded strings.
    # This is not a problem for round trip JSON, but it is a problem
    # when doing a round trip conversion from a message to a dict to a message.
    # See https://github.com/protocolbuffers/protobuf/issues/2679
    # and
    # https://developers.google.com/protocol-buffers/docs/proto3#json
    # for more details.
    class Squid(proto.Message):
        mass_kg = proto.Field(proto.INT64, number=1)
        length_cm = proto.Field(proto.UINT64, number=2)
        age_s = proto.Field(proto.FIXED64, number=3)
        depth_m = proto.Field(proto.SFIXED64, number=4)
        serial_num = proto.Field(proto.SINT64, number=5)

    s = Squid(mass_kg=10, length_cm=20, age_s=30, depth_m=40, serial_num=50)

    s_dict = Squid.to_dict(s)

    s2 = Squid(s_dict)

    assert s == s2

    # Double check that the conversion works with deeply nested messages.
    class Clam(proto.Message):
        class Shell(proto.Message):
            class Pearl(proto.Message):
                mass_kg = proto.Field(proto.INT64, number=1)

            pearl = proto.Field(Pearl, number=1)

        shell = proto.Field(Shell, number=1)

    c = Clam(shell=Clam.Shell(pearl=Clam.Shell.Pearl(mass_kg=10)))

    c_dict = Clam.to_dict(c)

    c2 = Clam(c_dict)

    assert c == c2



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import proto


def test_oneof():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1, oneof="bacon")
        baz = proto.Field(proto.STRING, number=2, oneof="bacon")

    foo = Foo(bar=42)
    assert foo.bar == 42
    assert not foo.baz
    foo.baz = "the answer"
    assert not foo.bar
    assert foo.baz == "the answer"


def test_multiple_oneofs():
    class Foo(proto.Message):
        bar = proto.Field(proto.INT32, number=1, oneof="spam")
        baz = proto.Field(proto.STRING, number=2, oneof="spam")
        bacon = proto.Field(proto.FLOAT, number=3, oneof="eggs")
        ham = proto.Field(proto.STRING, number=4, oneof="eggs")

    foo = Foo()
    foo.bar = 42
    foo.bacon = 42.0
    assert foo.bar == 42
    assert foo.bacon == 42.0
    assert not foo.baz
    assert not foo.ham
    foo.ham = "this one gets assigned"
    assert not foo.bacon
    assert foo.ham == "this one gets assigned"
    assert foo.bar == 42
    assert not foo.baz



# Copyright 2021, Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import enums_test


def test_total_ordering_w_same_enum_type():
    to_compare = enums_test.OneEnum.SOME_VALUE

    for item in enums_test.OneEnum:
        if item.value < to_compare.value:
            assert not to_compare == item
            assert to_compare != item
            assert not to_compare < item
            assert not to_compare <= item
            assert to_compare > item
            assert to_compare >= item
        elif item.value > to_compare.value:
            assert not to_compare == item
            assert to_compare != item
            assert to_compare < item
            assert to_compare <= item
            assert not to_compare > item
            assert not to_compare >= item
        else:  # item.value == to_compare.value:
            assert to_compare == item
            assert not to_compare != item
            assert not to_compare < item
            assert to_compare <= item
            assert not to_compare > item
            assert to_compare >= item


def test_total_ordering_w_other_enum_type():
    to_compare = enums_test.OneEnum.SOME_VALUE

    for item in enums_test.OtherEnum:
        assert not to_compare == item
        assert to_compare.SOME_VALUE != item
        with pytest.raises(TypeError):
            assert not to_compare < item
        with pytest.raises(TypeError):
            assert not to_compare <= item
        with pytest.raises(TypeError):
            assert not to_compare > item
        with pytest.raises(TypeError):
            assert not to_compare >= item


@pytest.mark.parametrize("int_val", range(-1, 3))
def test_total_ordering_w_int(int_val):
    to_compare = enums_test.OneEnum.SOME_VALUE

    if int_val < to_compare.value:
        assert not to_compare == int_val
        assert to_compare != int_val
        assert not to_compare < int_val
        assert not to_compare <= int_val
        assert to_compare > int_val
        assert to_compare >= int_val
    elif int_val > to_compare.value:
        assert not to_compare == int_val
        assert to_compare != int_val
        assert to_compare < int_val
        assert to_compare <= int_val
        assert not to_compare > int_val
        assert not to_compare >= int_val
    else:  # int_val == to_compare.value:
        assert to_compare == int_val
        assert not to_compare != int_val
        assert not to_compare < int_val
        assert to_compare <= int_val
        assert not to_compare > int_val
        assert to_compare >= int_val


def test_hashing():
    to_hash = enums_test.OneEnum.SOME_VALUE

    {to_hash: "testing"}  # no raise



# Copyright (C) 2020  Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import proto
from google.protobuf.json_format import MessageToJson, Parse, ParseError


def test_message_to_json():
    class Squid(proto.Message):
        mass_kg = proto.Field(proto.INT32, number=1)

    s = Squid(mass_kg=100)
    json = Squid.to_json(s)
    json = json.replace(" ", "").replace("\n", "")
    assert json == '{"massKg":100}'


def test_message_from_json():
    class Squid(proto.Message):
        mass_kg = proto.Field(proto.INT32, number=1)

    json = """{
    "massKg": 100
    }
    """

    s = Squid.from_json(json)
    assert s == Squid(mass_kg=100)


def test_message_json_round_trip():
    class Squid(proto.Message):
        mass_kg = proto.Field(proto.INT32, number=1)

    s = Squid(mass_kg=100)
    json = Squid.to_json(s)
    s2 = Squid.from_json(json)

    assert s == s2


def test_json_stringy_enums():
    class Squid(proto.Message):
        zone = proto.Field(proto.ENUM, number=1, enum="Zone")

    class Zone(proto.Enum):
        EPIPELAGIC = 0
        MESOPELAGIC = 1
        BATHYPELAGIC = 2
        ABYSSOPELAGIC = 3

    s1 = Squid(zone=Zone.MESOPELAGIC)
    json = (
        Squid.to_json(s1, use_integers_for_enums=False)
        .replace(" ", "")
        .replace("\n", "")
    )
    assert json == '{"zone":"MESOPELAGIC"}'

    s2 = Squid.from_json(json)
    assert s2.zone == s1.zone


def test_json_default_enums():
    class Squid(proto.Message):
        zone = proto.Field(proto.ENUM, number=1, enum="Zone")

    class Zone(proto.Enum):
        EPIPELAGIC = 0
        MESOPELAGIC = 1
        BATHYPELAGIC = 2
        ABYSSOPELAGIC = 3

    s = Squid()
    assert s.zone == Zone.EPIPELAGIC
    json1 = Squid.to_json(s).replace(" ", "").replace("\n", "")
    assert json1 == '{"zone":0}'

    json2 = (
        Squid.to_json(s, use_integers_for_enums=False)
        .replace(" ", "")
        .replace("\n", "")
    )
    assert json2 == '{"zone":"EPIPELAGIC"}'


def test_json_default_values():
    class Squid(proto.Message):
        mass_kg = proto.Field(proto.INT32, number=1)
        name = proto.Field(proto.STRING, number=2)

    s = Squid(name="Steve")
    json1 = (
        Squid.to_json(s, including_default_value_fields=False)
        .replace(" ", "")
        .replace("\n", "")
    )
    assert json1 == '{"name":"Steve"}'

    json2 = Squid.to_json(s).replace(" ", "").replace("\n", "")
    assert (
        json2 == '{"name":"Steve","massKg":0}' or json2 == '{"massKg":0,"name":"Steve"}'
    )

    s1 = Squid.from_json(json1)
    s2 = Squid.from_json(json2)
    assert s == s1 == s2


def test_json_unknown_field():
    # Note that 'lengthCm' is unknown in the local definition.
    # This could happen if the client is using an older proto definition
    # than the server.
    json_str = '{\n  "massKg": 20,\n  "lengthCm": 100\n}'

    class Octopus(proto.Message):
        mass_kg = proto.Field(proto.INT32, number=1)

    o = Octopus.from_json(json_str, ignore_unknown_fields=True)
    assert not hasattr(o, "length_cm")
    assert not hasattr(o, "lengthCm")

    # Don't permit unknown fields by default
    with pytest.raises(ParseError):
        o = Octopus.from_json(json_str)


def test_json_snake_case():
    class Squid(proto.Message):
        mass_kg = proto.Field(proto.INT32, number=1)

    json_str = '{\n  "mass_kg": 20\n}'
    s = Squid.from_json(json_str)

    assert s.mass_kg == 20

    assert Squid.to_json(s, preserving_proto_field_name=True) == json_str


def test_json_name():
    class Squid(proto.Message):
        massKg = proto.Field(proto.INT32, number=1, json_name="mass_in_kilograms")

    s = Squid(massKg=20)
    j = Squid.to_json(s)

    assert "mass_in_kilograms" in j

    s_two = Squid.from_json(j)

    assert s == s_two
